# Sack
# by Eresse <eresse@eresse.net>

# External Includes
require 'thread'

# Internal Includes
require 'sack/database/data'

# Sack Module
module Sack

	# Database Class:
	# Main Database Access Class - provides CRUD methods (and a little more) against a given Schema on a provided Database Connector and Connection String.
	class Database

		# Actions:
		# Allowed data methods
		ACTIONS = [
			:create_table,
			:count,
			:find,
			:find_by,
			:fetch,
			:fetch_by,
			:fetch_all,
			:create,
			:update,
			:update_by,
			:save,
			:delete,
		    :delete_by
		]

		# Construct:
		# Builds a *Database* on top of a backend _connector_ and _connstring_, set to operate on _schema_.
		# @param [Object] connector Generic database backend connector
		# @param [String] connstring Connector-specific connection string
		# @param [Hash] schema Schema definition - see README
		def initialize connector, connstring, schema

			# Set Connector & Connstring
			@connector = connector
			@connstring = connstring

			# Set Schema
			@schema = schema

			# Create Lock
			@lock = Mutex.new

			# Verify Schema
			verify_schema
		end

		# Open:
		# Opens a session on the database for yielding.
		def open

			# Lock DB
			@lock.synchronize do

				# Open Database
				db = @connector.open @connstring

				# Yield Block
				yield Data.new(@connector, db, @schema) if block_given?

				# Close Database
				db.close
			end
		end

		# Verify Schema:
		# Verifies the supplied schema against the actual database & re-creates it if necessary.
		def verify_schema
			open { |data| rebuild_schema data unless check_schema data }
		end

		# Check Schema:
		# Verifies the supplied schema against the actual database.
		# @param [Data] data Data access interface
		# @return [Object] True if schema is valid, False otherwise
		def check_schema data
			@schema.keys.inject(true) { |a, table| a && (data.exec "select count(*) from #{table};" rescue nil) }
		end

		# Rebuild Schema:
		# Re-creates the database according to the supplied schema.
		# @param [Data] data Data access interface
		def rebuild_schema data
			@schema.each { |table, fields| data.create_table table, fields rescue data.alter_table table, fields }
			raise "Rebuilding Schema Failed - Check Database & Access Permissions" unless check_schema data
		end

		# Method Missing:
		# Catches and routes database actions through data access interface.
		def method_missing name, *args

			# Check Action
			raise "Unknown action [#{name}]" unless ACTIONS.include? name

			# Open Database { Perform Action }
			open { |db| return db.send name, *args }
		end
	end
end
