# Sack
# by Eresse <eresse@eresse.net>

# External Includes
require 'aromat'

# Internal Includes
require 'sack/database/ftypes'
require 'sack/database/statement'
require 'sack/database/generator'
require 'sack/database/sanitizer'

# Sack Module
module Sack

	# Database Class
	class Database

		# Data Class:
		# Internal Class presented as Data Access Layer when yielding in Database.open.
		class Data

			# Construct:
			# Builds a *Data* object around a _db_ instance, set to operate on _schema_.
			# @param [Object] db Database instance obtained by DBMS file
			# @param [Hash] schema Schema definition - see README
			def initialize connector, db, schema
				@connector = connector
				@db = db
				@schema = schema
			end

			# Create Table:
			# Creates a table _name_ defined by the _fields_ hash.
			# @param [Symbol] name Table name
			# @param [Hash] fields A hash mapping of field names to type definition arrays (from _FTYPES_)
			def create_table name, fields
				fq = fields.collect { |fname, ftype| "#{Sanitizer.field_name fname} #{ftype.respond_to?(:each) ? ftype.collect { |e| FTYPES[Sanitizer.ftype e] }.join(' ') : FTYPES[Sanitizer.ftype ftype]}" }.join ', '
				@connector.exec @db, "create table #{Sanitizer.table @schema, name} (#{fq});"
			end

			# Alter Table:
			# Adds missing fields to an already-existing table (ONLY adds, does not remove or modify existing fields).
			# @param [Symbol] name Table name
			# @param [Hash] fields A hash mapping of field names to type definition arrays (from _FTYPES_)
			def alter_table name, fields
				fields.each { |fname, ftype| @connector.exec @db, "alter table #{Sanitizer.table @schema, name} add #{Sanitizer.field_name fname} #{ftype.respond_to?(:each) ? ftype.collect { |e| FTYPES[Sanitizer.ftype e] }.join(' ') : FTYPES[Sanitizer.ftype ftype]};" rescue nil }
			end

			# Count:
			# Counts the number of rows for a given table.
			# @param [Symbol] table Table name
			# @return [Fixnum] The number of rows present in the given _table_
			def count table
				@connector.exec(@db, "select count(*) from #{Sanitizer.table @schema, table};")[0][0]
			end

			# Find:
			# Fetches the first matching row from a given _table_ by _id_.
			def find table, id
				fetch(table, id).try :first
			end

			# Find By Field:
			# Fetches the first matching row from a given _table_ where _field_ matches _val_.
			def find_by table, field, val
				fetch_by(table, field, val).try :first
			end

			# Fetch:
			# Fetches rows from a given _table_ by _id_.
			# @param [Symbol] table Table name
			# @param [Object] id ID on which to filter
			# @return [Array] An array of Hashes, each representing one row
			def fetch table, id
				hash_res(table.to_sym, @connector.exec(@db, Statement.prep("select * from #{Sanitizer.table @schema, table} where id = ?;", [id])))
			end

			# Fetch By Field:
			# Fetches rows from a given _table_ where _field_ matches _val_.
			# @param [Symbol] table Table name
			# @param [Symbol] field Field name
			# @param [Object] val Field value
			# @return [Array] An array of Hashes, each representing one row
			def fetch_by table, field, val
				hash_res(table.to_sym, @connector.exec(@db, Statement.prep("select * from #{Sanitizer.table @schema, table} where #{Sanitizer.field @schema, table, field} = ?;", [val])))
			end

			# Fetch All:
			# Fetches all rows from a given _table_.
			# @param [Symbol] table Table name
			# @return [Array] An array of Hashes, each representing one row
			def fetch_all table
				hash_res(table.to_sym, @connector.exec(@db, "select * from #{Sanitizer.table @schema, table};"))
			end

			# Create:
			# Inserts _fields_ as a row into a given _table_.
			# @param [Symbol] table Table name
			# @param [Hash] fields Fields to be inserted
			def create table, fields
				@connector.exec @db, Statement.prep("insert into #{Sanitizer.table @schema, table} (#{Generator.fields @schema, table, fields}) values (#{Generator.marks fields});", Generator.values(fields.values))
			end

			# Update:
			# Updates _fields_ in rows identified by _id_ in a given _table_.
			# @param [Symbol] table Table name
			# @param [Object] id ID on which to filter
			# @param [Hash] fields Fields to be updated
			def update table, id, fields
				@connector.exec @db, Statement.prep("update #{Sanitizer.table @schema, table} set #{Generator.update_marks @schema, table, fields} where id = ?;", [Generator.values(fields.values), id].flatten)
			end

			# Update By Field:
			# Updates _fields_ in rows where _field_ matches _val_.
			# @param [Symbol] table Table name
			# @param [Symbol] field Field name
			# @param [Object] val Field value
			# @param [Hash] fields Fields to be updated
			def update_by table, field, val, fields
				@connector.exec @db, Statement.prep("update #{Sanitizer.table @schema, table} set #{Generator.update_marks @schema, table, fields} where #{Sanitizer.field @schema, table, field} = ?;", [Generator.values(fields.values), val].flatten)
			end

			# Save:
			# Creates or updates _fields_ in a given _table_, depending on the presence of an _id_.
			# @param [Symbol] table Table name
			# @param [Hash] fields Fields to be inserted / updated
			def save table, fields
				if fields[:id]
					update table, fields.clone.delete(:id), fields
				else
					create table, fields
				end
			end

			# Destroy:
			# Removes rows from a given _table_ where _field_ matches _val_.
			# @param [Symbol] table Table name
			# @param [Symbol] field Field name
			# @param [Object] val Field value
			def delete_by table, field, val
				@connector.exec @db, Statement.prep("delete from #{Sanitizer.table @schema, table} where #{Sanitizer.field @schema, table, field} = ?;", [val])
			end

			# Destroy By Field:
			# Removes rows identified by _id_ from a given _table_.
			# @param [Symbol] table Table name
			# @param [Object] id ID of rows to be removed
			def delete table, id
				@connector.exec @db, Statement.prep("delete from #{Sanitizer.table @schema, table} where id = ?;", [id])
			end

			# Execute statement:
			# Generic method to execute any SQL statement.
			# @param [String] q Statement to be executed
			# @param [Array] params Statement parameters
			# @return [Object] Whatever the statement returned
			def exec q, params = []
				@connector.exec @db, Statement.prep(q, params)
			end

			# Pull Results into Hash:
			# Converts rows returned by DBMS into Hashes matching the provided schema.
			# @param [Symbol] table Table name
			# @param [Array] x Results returned by DBMS
			# @return [Array] An array of Hashes, each representing a single row
			def hash_res table, x
				x
					.collect { |r| @schema[table].keys.each_with_index.inject([]) { |a, e| a + [e[0], r[e[1]]] } }
					.collect { |r| Hash[*r] }
					.sym_keys rescue nil
			end
		end
	end
end
