# Sack
# by Eresse <eresse@eresse.net>

# Internal Includes
require 'sack/database/sanitizer'

# Sack Module
module Sack

	# Database Class
	class Database

		# Generator Module:
		# Provides SQL generator methods for building queries.
		module Generator

			# Generate Field Name List:
			# Builds a list of field names present in _field_hash_, separated by commas.
			# @param [Hash] schema Database schema
			# @param [Symbol] table Table name
			# @param [Hash] field_hash Hash containing field names as keys
			# @return [String] A comma-separated list of field names
			def self.fields schema, table, field_hash
				field_hash.keys.collect { |k| Sanitizer.field schema, table, k }.join ', '
			end

			# Value Marks:
			# Builds a list of field markers ('?') for fields present in _field_hash_, separated by commas.
			# @param [Hash] field_hash Hash containing fields
			# @return [String] A comma-separated list of field markers ('?')
			def self.marks field_hash
				(['?'] * field_hash.size).join ', '
			end

			# Update Marks:
			# Builds a list of field value markers ('field = ?') for fields present in _field_hash_, separated by commas.
			# @param [Hash] schema Database schema
			# @param [Symbol] table Table name
			# @param [Hash] field_hash Hash containing field names as keys
			# @return [String] A comma-separated list of field value markers ('field = ?')
			def self.update_marks schema, table, field_hash
				field_hash.collect { |k, v| "#{Sanitizer.field schema, table, k} = #{v ? '?' : 'NULL'}" }.join ', '
			end

			# Values:
			# Replaces all symbols in _vals_ by their stringified values.
			# @param [Array] vals Array of value objects
			# @return [Array] A copy of _vals_ where every symbol has been replaced by its string representation
			def self.values vals
				vals.collect { |v| v.nil? ? nil : (v.is_a?(Symbol) ? v.to_s : v) }.reject { |e| e.nil? }
			end
		end
	end
end
