# Sack
# by Eresse <eresse@eresse.net>

# External Includes
require 'aromat'

# Internal Includes
require 'sack/database/ftypes'
require 'sack/database/sanitizer'
require 'sack/database/model/data'
require 'sack/database/model/validation'
require 'sack/database/model/relationships'

# Sack Module
module Sack

	# Database Class
	class Database

		# Model Module:
		# Provides abstractions for defining database models.
		module Model

			# Included:
			# Inject stuff when included.
			# @param [Object] base Whatever we've been included into
			def self.included base

				# Set Model
				base.instance_variable_set '@model', base

				# Link to parent Data Model Root
				base.instance_variable_set '@model_root', base.mod_parent

				# Set Model Name
				base.instance_variable_set '@model_name', base.mod_name

				# Extend Class Methods
				base.extend ClassMethods

				# Extend with Data Access Methods
				base.extend Data

				# Include Validation
				base.include Validation

				# Include Relationships
				base.include Relationships
			end

			# Class Methods:
			# Collection of methods to be injected into anything that includes this module.
			module ClassMethods

				# Get / Set Table Name:
				# Determines the default table name through introspection, or overrides the default table name (if _name_ is provided).
				# @param [Symbol] name Override table name with this
				# @return [Symbol] The table name
				def table_name name = nil

					# Introspect Default Name
					@table_name ||= self.mod_name.snakecase

					# Override with custom name
					@table_name = name if name

					@table_name.to_sym
				end

				# Set Field:
				# Configures a field on the current Model.
				# @param [Hash] options Options defining the field - see README
				def field options

					# Internalise Options (so we can mess it up)
					options = options.clone

					# Pull Name
					name, ftype = options.first
					options.delete name

					# Check Primary Type
					raise "Invalid Field Primary Type [#{ftype.first}] for [#{name}] in #{mod_name}" unless FTYPES_CLASSES.include? ftype.first

					# Collect Validation Rules
					@fields ||= {}
					@fields[name] ||= {}
					@fields[name][:ftype] = ftype
					@fields[name][:rules] = options
				end

				# Get Fields:
				# Simply returns the model's field map.
				# @return [Hash] The field map for the current model
				def fields
					@fields
				end

				# Get Field Schema:
				# Builds a schema for the current model.
				# @return [Hash] The current model's schema
				def field_schema
					Hash[*(@fields.inject([]) { |a, e| a << e[0] << e[1][:ftype] })]
				end
			end
		end
	end
end
