# Sack
# by Eresse <eresse@eresse.net>

# External Includes
require 'aromat'

# Internal Includes
require 'sack/database/ftypes'

# Sack Module
module Sack

	# Database Class
	class Database

		# Model Module
		module Model

			# Validation Module:
			# Provides a simple entity validation framework.
			module Validation

				# Validation Exception
				class ValidationException < Exception
					attr_reader :errors
					def initialize msg, errors
						super msg
						@errors = errors
					end
				end

				# Included:
				# Inject Validation Methods when included.
				# @param [Object] base Whatever we've been included into
				def self.included base
					base.extend ClassMethods
				end

				# Class Methods:
				# Collection of methods to be injected into anything that includes this module.
				module ClassMethods

					# Is Valid:
					# Verifies the validity of a given entity (_data_) against this Model.
					# @param [Database] db Database instance (Sack::Database)
					# @param [Hash] data Entity data
					def is_valid? db, data, errors = []

						# Run through Model's Field Schema
						fields.inject(true) do |a, e|

							# Acquire Field Info
							name = e[0]
							info = e[1]
							ftype = info[:ftype]
							rules = info[:rules]
							val = data[name]

							# Handle Required
							if rules[:required]
								r = val.try :is_a?, FTYPES_CLASSES[ftype.first]
								a &&= r
								errors << "Required field [#{name}] is missing" unless r
							end

							# Handle Unique
							if rules[:unique]
								r = is_uniq? db, data, name, val, rules[:unique]
								a &&= r
								errors << "Field [#{name}] has non-unique value [#{val}]#{rules[:unique].is_a?(Array) ? " in scope [#{rules[:unique].join ', '}]" : ''}" unless r
							end

							# Handle Set Length
							if rules[:length]
								r = val.try(:length).try :==, rules[:length]
								a &&= r
								errors << "Field [#{name}] has invalid length (#{val.try(:length)} / #{rules[:length]})" unless r
							end

							# Handle Minimum Length
							if rules[:min_length]
								r = val.try(:length).try :>=, rules[:min_length]
								a &&= r
								errors << "Field [#{name}] has invalid length (#{val.try(:length)} / Min: #{rules[:min_length]})" unless r
							end

							# Handle Maximum Length
							if rules[:max_length]
								r = val.try(:length).try :<=, rules[:max_length]
								a &&= r
								errors << "Field [#{name}] has invalid length (#{val.try(:length)} / Max: #{rules[:max_length]})" unless r
							end

							# Handle Regex
							if rules[:regex]
								r = rules[:regex] =~ val
								a &&= r
								errors << "Field [#{name}] doesn't match allowed pattern (#{rules[:regex].inspect})" unless r
							end

							# Handle Custom
							if rules[:validate]
								size = errors.size
								r = send rules[:validate], db, data, name, val, rules, errors
								a &&= r
								errors << "Field [#{name}] is invalid" unless r || (size != errors.size)
							end

							# Don't leak shit
							!!a
						end
					end

					# Is Unique:
					# Verifies the unicity of a given field (_name_) value (_val_) on an entity (_data_), possibly within a given _scope_.
					# @param [Database] db Database instance (Sack::Database)
					# @param [Hash] data Entity data
					# @param [Symbol] name Field name
					# @param [Object] val Field value
					# @param [Object] scope Either True (if absolute - no scope) or an Array of field names defining the scope
					def is_uniq? db, data, name, val, scope

						# Fetch all other rows with field [name] equal to [val]
						others = db.fetch_by table_name, name, val
						others.reject! { |o| o[:id] == data[:id] } if data[:id]

						# If scopeless (absolutely unique), check empty set and return
						return others.empty? unless scope.is_a? Array

						# Check Unique throughout Scope
						scope.each { |f| others.reject! { |o| o[f] != data[f] } }
						others.empty?
					end
				end
			end
		end
	end
end
