# Sack
# by Eresse <eresse@eresse.net>

# Internal Includes
require 'sack/database/ftypes'

# Sack Module
module Sack

	# Database Class
	class Database

		# Santizer Module:
		# Provides Table and Field name sanitization methods.
		module Sanitizer

			# Generic Field Name Regex
			FIELD_NAME_REX = /^[0-9a-z_.-]+$/

			# Sanitize Table Name:
			# Raises an exception if _name_ is not a valid table in _schema_.
			# @param [Hash] schema Database schema
			# @param [Symbol] name Table name to sanitize
			# @return [Symbol] Table name if sanitization passed
			def self.table schema, name
				raise "Illegal table name [#{name}]" unless (name.to_sym.to_s == name.to_s) && schema.has_key?(name.to_sym)
				name
			end

			# Sanitize Table Field Name:
			# Raises an exception if _table_ or _field_ are not valid according to _schema_.
			# @param [Hash] schema Database schema
			# @param [Symbol] table Table name
			# @param [Symbol] field Field name
			# @return [Symbol] Field name if sanitization passed
			def self.field schema, table, field
				table schema, table
				raise "Illegal field [#{field}] for table [#{table}]" unless (field.to_sym.to_s == field.to_s) && schema[table.to_sym].has_key?(field.to_sym)
				field
			end

			# Sanitize Generic Field Name:
			# Raises an exception if _name_ contains invalid characters (defined in _FIELD_NAME_REX_).
			# @param [Symbol] name Field name
			# @return [Symbol] Field name if sanitization passed
			def self.field_name name
				raise "Illegal field name [#{name}]" unless FIELD_NAME_REX =~ name
				name
			end

			# Sanitize Field Types
			# Raises an exception if _t_ is not an allowed Field Type (defined in _FTYPES_).
			# @param [Symbol] t Field type symbol (from _FTYPES_)
			# @return [Symbol] Field type if sanitization passed
			def self.ftype t
				raise "Illegal field type [#{t}]" unless FTYPES.keys.include? t
				t
			end

			# Sanitize Field Value:
			# Escapes single-quotes inside field values.
			# @param [Object] v Field value
			# @return [Object] The supplied value, with single quotes escaped if it's a String.
			def self.value v
				return v unless v.is_a? String
				drop_nonprintascii(v).gsub("'") { "''" }
			end

			# Drop Non-Print-ASCII:
			# Removes all non-printable-ASCII characters from a String.
			# @param [String] s Input string
			# @return [String] The provided string, stripped of any non-printable-ASCII text
			def self.drop_nonprintascii s
				s.bytes.select { |b| (b >= 0x20) && (b <= 0x7e) }.inject('') { |a, e| a + e.chr }
			end
		end
	end
end
